@@ -104,19 +104,13 @@ class SmartDriver(SeleniumDriverCore):
                     'source': source,
                     'api_key': self.api_key,
                     'label': element_name,
+                    'do_exact_match_first': self.do_exact_match_first,
                 }
-                classify_url = self.url + '/detect'
                 start = time.time()
-                for _ in range(2):
-                    try:
-                        r = requests.post(classify_url, json=data, verify=False, timeout=10)
-                        break
-                    except Exception:
-                        log.error('error during detect')
+                response = self.make_json_post_request('/detect', data, self.detect_timeout_message, self.detect_timeout)
                 end = time.time()
                 if self.debug:
                     print(f'Classify time: {end - start}')
-                response = r.json()
                 if not response['success']:
                     classification_error_msg = response['message'].replace(self.default_prod_url, self.url)
                     raise Exception(classification_error_msg)
