@@ -30,6 +30,8 @@ from selenium.common.exceptions import StaleElementReferenceException
 
 from . import __version__ as base_version
 from .utils.selenium_core import SeleniumDriverCore
+from .utils.local_classify import LocalClassifier
+
 
 requests.packages.urllib3.disable_warnings()
 
@@ -51,6 +53,11 @@ class SmartDriver(SeleniumDriverCore):
         self_attrs = dir(self)
         # Disable warnings
         requests.packages.urllib3.disable_warnings()
+        self.local_classify = None
+        if self.do_local_caching:
+            self.local_classify = LocalClassifier(self.url, self.api_key, template_match_threshold=self.local_match_threshold)
+            for element_name in self.element_names_in_tc:
+                self.local_classify.cache_templates_for_element(element_name)
 
         warnings.filterwarnings("ignore", category=DeprecationWarning)
 
@@ -66,9 +73,7 @@ class SmartDriver(SeleniumDriverCore):
                 elems = json.loads(raw_content)
                 for e in elems:
                     if e.get('label') == element_name:
-                        real_elem = self._match_bounding_box_to_selenium_element(e, multiplier=self.multiplier, offset=offset)
-                        parent_elem = real_elem.parent
-                        element = ai_elem(parent_elem, real_elem, e, self.driver, self.multiplier)
+                        element = self._get_matched_elem(e)
                         return element
                 return None
             except Exception as err:
@@ -115,6 +120,22 @@ class SmartDriver(SeleniumDriverCore):
             screenshotBase64 = self.driver.get_screenshot_as_base64()
         return screenshotBase64
 
+    def _get_matched_elem(self, element_box):
+        if self.use_cdp:
+            parent_elem = None
+            real_elem = element_box
+        else:
+            try:
+                real_elem = self._match_bounding_box_to_selenium_element(element_box, multiplier=self.multiplier)
+                parent_elem = real_elem.parent
+            except NoElementFoundException as e:
+                log.warning(
+                    "Could not find a regular Selenium element at coordinates, switching to AIElement, some attributes might be missing.")
+                parent_elem = None
+                real_elem = element_box
+        element = ai_elem(parent_elem, real_elem, element_box, self.driver, self.multiplier, self.use_cdp)
+        return element
+
 
     def process_exist_screenshot_response(self, element_name, key, msg, tresp_data, is_backup=False):
         if self.debug:
@@ -122,15 +143,7 @@ class SmartDriver(SeleniumDriverCore):
         element_box = tresp_data['predicted_element']
         expected_offset = tresp_data['page_offset']
         element_box['y'] += expected_offset
-
-        if self.use_cdp:
-            parent_elem = None
-            real_elem = element_box
-        else:
-            real_elem = self._match_bounding_box_to_selenium_element(element_box, multiplier=self.multiplier)
-            parent_elem = real_elem.parent
-        element = ai_elem(parent_elem, real_elem, element_box, self.driver,
-                          self.multiplier)
+        element = self._get_matched_elem(element_box)
         try:
             real_xpath = self._generate_xpath(element)
             self._send_warning(real_xpath, element_name, key, tresp_data, is_backup=is_backup)
@@ -141,16 +154,25 @@ class SmartDriver(SeleniumDriverCore):
     def _classify(self, element_name, is_backup=True):
         msg = ''
         if self.test_case_creation_mode:
+            if self.do_local_caching:
+                screenshotBase64 = self._get_screenshot()
+                self.last_test_case_screenshot_uuid = self.get_screenshot_hash(screenshotBase64)
+
+                element_box = self.local_classify.classify_element(element_name, screenshotBase64)
+                if element_box is not None:
+                    msg = 'Found using local cache'
+                    element = self._get_matched_elem(element_box)
+                    return element, self.last_test_case_screenshot_uuid, msg
+                else:
+                    msg = 'Not found using local cache'
+                needs_reload = False
+                has_training_data = True
+
+
             self._test_case_upload_screenshot(element_name)
             element_box, needs_reload, has_training_data = self._test_case_get_box(element_name)
             if element_box:
-                if self.use_cdp:
-                    parent_elem = None
-                    real_elem = element_box
-                else:
-                    real_elem = self._match_bounding_box_to_selenium_element(element_box, multiplier=self.multiplier)
-                    parent_elem = real_elem.parent
-                element = ai_elem(parent_elem, real_elem, element_box, self.driver, self.multiplier)
+                element = self._get_matched_elem(element_box)
                 return element, self.last_test_case_screenshot_uuid, msg
             else:
                 if has_training_data:
@@ -159,13 +181,7 @@ class SmartDriver(SeleniumDriverCore):
                         element_box = response['predicted_element']
                         self.page_offset = self.driver.execute_script('return window.pageYOffset')
                         element_box['y'] += self.page_offset * self.multiplier
-                        if self.use_cdp:
-                            parent_elem = None
-                            real_elem = element_box
-                        else:
-                            real_elem = self._match_bounding_box_to_selenium_element(element_box, multiplier=self.multiplier)
-                            parent_elem = real_elem.parent
-                        element = ai_elem(parent_elem, real_elem, element_box, self.driver, self.multiplier)
+                        element = self._get_matched_elem(element_box)
                         return element, self.last_test_case_screenshot_uuid, msg
                     else:
                         # scroll back up
@@ -178,15 +194,7 @@ class SmartDriver(SeleniumDriverCore):
                 while True:
                     element_box, needs_reload, _ = self._test_case_get_box(element_name, event_id=event_id)
                     if element_box is not None:
-                        if self.use_cdp:
-                            parent_elem = None
-                            real_elem = element_box
-                        else:
-                            real_elem = self._match_bounding_box_to_selenium_element(element_box,
-                                                                                     multiplier=self.multiplier)
-                            parent_elem = real_elem.parent
-                        element = ai_elem(parent_elem, real_elem, element_box, self.driver,
-                                          self.multiplier)
+                        element = self._get_matched_elem(element_box)
                         return element, self.last_test_case_screenshot_uuid, msg
 
                     if needs_reload:
@@ -200,6 +208,14 @@ class SmartDriver(SeleniumDriverCore):
             ## Get screenshot & page source
             screenshotBase64 = self._get_screenshot()
             key = self.get_screenshot_hash(screenshotBase64)
+
+            if self.do_local_caching:
+                local_pred = self.local_classify.classify_element(element_name, screenshotBase64)
+                if local_pred is not None:
+                    msg = 'Found using local cache'
+                    element = self._get_matched_elem(local_pred)
+                    return element, key, msg
+
             resp_data = self._check_screenshot_exists(key, element_name)
 
             if resp_data['success'] and 'predicted_element' in resp_data and resp_data['predicted_element'] is not None:
@@ -252,19 +268,13 @@ class SmartDriver(SeleniumDriverCore):
                 element_box = response['predicted_element']
                 self.page_offset = self.driver.execute_script('return window.pageYOffset')
                 element_box['y'] += self.page_offset * self.multiplier
-
-                if self.use_cdp:
-                    parent_elem = None
-                    real_elem = element_box
-                else:
-                    real_elem = self._match_bounding_box_to_selenium_element(element_box, multiplier=self.multiplier)
-                    parent_elem = real_elem.parent
-                element = ai_elem(parent_elem, real_elem, element_box, self.driver, self.multiplier)
-                try:
-                    real_xpath = self._generate_xpath(element)
-                    self._send_warning(real_xpath, element_name, key, response, is_backup=is_backup)
-                except Exception as e:
-                    log.debug(e)
+                element = self._get_matched_elem(element_box)
+                if element._is_real_elem:
+                    try:
+                        real_xpath = self._generate_xpath(element)
+                        self._send_warning(real_xpath, element_name, key, response, is_backup=is_backup)
+                    except Exception as e:
+                        log.debug(e)
             except Exception as e:
                 log.error(e)
                 logging.exception('exception during classification')
@@ -407,8 +417,9 @@ class SmartDriver(SeleniumDriverCore):
             return False
 
 class ai_elem(webdriver.remote.webelement.WebElement):
-    def __init__(self, parent, source_elem, elem, driver, multiplier=1.0):
+    def __init__(self, parent, source_elem, elem, driver, multiplier=1.0, use_cdp=False):
         self._is_real_elem = False
+        self._use_cdp = use_cdp
         if not isinstance(source_elem, dict):
             # We need to also pass the _w3c flag otherwise the get_attribute for thing like html or text is messed up
             if old_selenium:
@@ -467,7 +478,7 @@ class ai_elem(webdriver.remote.webelement.WebElement):
 
 
     def click(self, js_click=False, use_cdp=False):
-        if self._is_real_elem == True and use_cdp == False:
+        if self._is_real_elem == True and use_cdp == False and self._use_cdp == False:
             if not js_click:
                 webdriver.remote.webelement.WebElement.click(self)
             else:
@@ -493,3 +504,7 @@ class ai_elem(webdriver.remote.webelement.WebElement):
 
 class NoElementFoundException(Exception):
     pass
+
+from selenium.webdriver.common.by import By
+class By(By):
+    AI = 'ai'
\ No newline at end of file
