@@ -64,6 +64,9 @@ class SeleniumDriverCore(object):
         self.detect_timeout = initialization_options.get('detect_timeout', 60)
         self.misc_timeout = initialization_options.get('misc_timeout', 10)
         self.do_exact_match_first = initialization_options.get('do_exact_match_first', False)
+        self.do_local_caching = initialization_options.get('local_caching', False)
+        self.local_match_threshold = initialization_options.get('local_match_threshold', 0.998)
+        self.element_names_in_tc = []
         self.exact_match_first_threshold = initialization_options.get('exact_match_first_threshold', 0.999)
 
         self.url = initialization_options.get('server_url',
@@ -402,7 +405,7 @@ class SeleniumDriverCore(object):
 
         if el is None:
             print(msg)
-            raise Exception(msg)
+            raise NoElementFoundException(msg)
         return el
 
     def _generic_find_method(self, find_method, element_name, *args):
@@ -411,7 +414,10 @@ class SeleniumDriverCore(object):
 
         # Run the standard selector
         try:
-            driver_element = find_method(*args)
+            if len(args) == 2 and args[0] == "ai":
+                driver_element = self.find_by_ai(args[1])
+            else:
+                driver_element = find_method(*args)
             if driver_element:
                 try:
                     key = self._upload_screenshot_if_necessary(element_name, driver_element=driver_element)
@@ -445,11 +451,17 @@ class SeleniumDriverCore(object):
                 'test_case_name': self.test_case_uuid,
                 'automation_name': self.automation_name}
         try:
-            res = requests.post(self.url + '/ping', json=data, timeout=3, verify=False)
+            res = requests.post(self.url + '/ping', json=data, timeout=10, verify=False)
             res = res.json()
             if not res['success']:
                 log.error(res['message'])
-        except Exception:
+            else:
+                if 'labels' not in res:
+                    log.info('Local matching disabled, no labels found during check_in')
+                log.debug(f'Got following elements from ping {res["labels"]}')
+                self.element_names_in_tc = res['labels']
+        except Exception as e:
+            log.debug(f'Failed to checkin, {e}')
             pass
 
     def make_json_post_request(self, route, data, timeout_error_message, timeout_variable, generic_error_message=None, tries=3):
@@ -460,6 +472,7 @@ class SeleniumDriverCore(object):
             generic_error_message = 'Error making request to ' + route
         for _ in range(tries):
             try:
+                log.debug(f'Posting to {route}')
                 url = self.url.rstrip('/') + route
                 res = requests.post(url, json=data, verify=False, timeout=local_timeout).json()
                 break
